import matplotlib as mpl
from matplotlib.colors import ListedColormap

__all__ = []

PALETTES = dict(
    wong=[
        '#000000',
        '#E69F00',
        '#50B4E9',
        '#009E73',
        '#F0E442',
        '#0072B2',
        '#D55E00',
        '#CC79A7',
    ],
    pantone=[
        '#9BB7D4',
        '#C74375',
        '#BF1932',
        '#7BC4C4',
        '#E2583E',
        '#53B0AE',
        '#DECDBE',
        '#9B1B30',
        '#5A5B9F',
        '#F0C05A',
        '#45B5AA',
        '#D94F70',
        '#DD4124',
        '#009473',
        '#B163A3',
        '#955251',
        '#F7CAC9',
        '#92A8D1',
        '#88B04B',
        '#5F4B8B',
        '#FF6F61',
        '#0F4C81',
        '#939597',
        '#F5DF4D',
        '#6667AB',
        '#BB2649',
    ],
)


def register_colormap(name: str, cmap: ListedColormap) -> None:
    try:
        if name not in mpl.colormaps:
            mpl.colormaps.register(cmap, name=name)
    except AttributeError:
        mpl.cm.register_cmap(name, cmap)


for _name, _lut in PALETTES.items():

    _cmap = ListedColormap(_lut, _name)
    locals()[_name] = _cmap

    _cmap_r = ListedColormap(_lut[::-1], _name + '_r')
    locals()[_name + '_r'] = _cmap_r

    register_colormap(_name, _cmap)
    register_colormap(_name + '_r', _cmap_r)


del PALETTES, register_colormap



#%%
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import gaussian_kde

from ..colormap.palette import dimmer
from ..typing.generics import *
from .scalers import MinMaxScaler, OneScaler
from .utils import ensure_fit

#%%
__all__ = ['dispersion_plot']

#%%
class KDE:
    def __init__(
        self,
        *,
        bw_method: BW_TYPES = 'scott',
        scale: bool = True,
    ):

        self.bw_method = bw_method
        self.scaler = MinMaxScaler() if scale else OneScaler()
        self.isfit = False
        self.name = 'KDE'

    def fit(self, X: ARRAY):
        self._kde = gaussian_kde(
            self.scaler.fit_transform(X), bw_method=self.bw_method
        )
        self.isfit = True
        return self

    @ensure_fit
    def __call__(self, X: ARRAY) -> np.ndarray:
        t = self.scaler.transform(X)
        return np.exp(self._kde.logpdf(t))


#%%
def dispersion_plot(
    x: ARRAY,
    y: ARRAY,
    *,
    jitter: float = 0.3,
    cmap: str = 'jet',
    ax: AXES = None,
    ybounds: BOUNDS = [(0, 1)],
) -> AXES:
    """Dispersion + Error bar + KDE plot for continous data with multiple groups.

    Params:
        x: A n-shape array indicating the groups
        y: A nxm-shape array with the value for each group
        jitter: The amplitude of dispersion of the data
        cmap: An available matplotlib colormap name
        ax: An axes to be plot
        ybounds: A list of bounds to restrict y data whithin

    Returns:
        The axes where the data was plot

    Examples:
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>>
        >>> from zeta.plot import dispersion_plot
        >>>
        >>> N = 5
        >>> X = np.arange(N)
        >>> Y = np.random.rand(N, 250)
        >>>
        >>> dispersion_plot(x=X, y=Y, cmap='pantone')
        >>> plt.show()

        ![dispersion](dispersion.png){width="300" .center}
    """
    assert (
        len(ybounds) == len(x) or len(ybounds) == 1
    ), f'ybounds should be 1 or the same size to x. Got {len(ybounds)}.'
    if len(ybounds) == 1:
        ybounds *= len(x)

    if ax is None:
        ax = plt.gca()

    for x_, y_, bounds_ in zip(x, y, ybounds):
        t = np.linspace(bounds_[0], bounds_[1], 100)
        kde = KDE().fit(y_)
        prob = kde(t)

        x1 = x_ - (prob / prob.max()) * jitter
        x2 = np.full_like(x1, x_)

        color = plt.get_cmap(cmap)(x_ / np.max(x))
        lighter, darker = dimmer(color, 1.05), dimmer(color, 0.7)

        ax.fill_betweenx(t, x1, x2, facecolor=lighter, alpha=0.7)
        ax.scatter(
            x_ + np.random.rand(len(y_)) * jitter,
            y_,
            alpha=0.7,
            edgecolors=lighter,
            facecolor='none',
        )
        ax.errorbar(x_, np.mean(y_), yerr=np.std(y), capsize=5, color=darker)

    return ax


#%%
def cocurrence_plot(x, y, *, group, cmap, ax=None):
    ...



import matplotlib as mpl
from cycler import cycler

__all__ = []

context = {
    'lines.marker': 's',
    'lines.linewidth': 1,
    'lines.markersize': 4,
    'figure.dpi': 300,
    'figure.titlesize': 10,
    'figure.labelsize': 8,
    'figure.figsize': (7.09, 3.14),
    'font.size': 8,
    'axes.prop_cycle': cycler('color', mpl.colormaps['pantone'].colors),
    'axes.grid': True,
    'grid.linestyle': 'dashed',
    'grid.alpha': 0.4,
}

mpl.rcParams.update(context)

del context, cycler, mpl



from colorsys import hls_to_rgb, rgb_to_hls

import numpy as np
from matplotlib import colors
from matplotlib.colors import LinearSegmentedColormap

from ..typing.generics import *

__all__ = ['blend_palette']


def _color2rgb(color: COLOR, format: COLOR_FORMAT):
    if format == 'hls':
        color = hls_to_rgb(*color)
    return colors.to_rgb(color)


def blend_palette(
    colors: COLORS_SEQ,
    format: COLOR_FORMAT = 'rgb',
) -> LinearSegmentedColormap:
    """A function that generates an interpolated colormap from
    the input sequence.

    Args:
        colors: A sequence of colors
        format: The format of the input, hex, rgb, hls

    Returns:
        A colormap from the inputed colors

    Examples:

    """
    colors = [_color2rgb(color, format) for color in colors]
    name = 'blend'
    pal = LinearSegmentedColormap.from_list(name, colors)
    return pal


def dimmer(color: COLOR, p: float):
    rgb = colors.to_rgb(color)
    hls = list(rgb_to_hls(*rgb))
    hls[1] = np.clip(hls[1] * p, 0, 1)
    return hls_to_rgb(*hls)



from abc import ABC, abstractmethod

import numpy as np

from ..typing.generics import *
from .utils import ensure_fit


class Scaler(ABC):
    def __init__(self, name: str):
        self.isfit = False
        self.name = name

    @abstractmethod
    def fit(self):
        ...

    @abstractmethod
    def transform(self):
        ...

    def fit_transform(self, X: ARRAY) -> ARRAY:
        return self.fit(X).transform(X)

    @abstractmethod
    def inverse_transform(self):
        ...


class MinMaxScaler(Scaler):
    def __init__(self):
        super().__init__(name='MinMaxScaler')

    def fit(self, X: ARRAY):
        self.min = np.nanmin(X)
        self.max = np.nanmax(X)
        self.scale = 1 / (self.max - self.min)
        self.x_std = X - self.min
        self.scaled = self.x_std * (self.max - self.min) + self.min

        self.isfit = True
        return self

    @ensure_fit
    def transform(self, X: ARRAY):
        return (X - self.min) * self.scale

    def inverse_transform(self, X):
        return X / self.scale + self.min


class OneScaler(Scaler):
    def __init__(self):
        super().__init__(name='OneScaler')

    def fit(self, X: ARRAY):
        self.isfit = True
        return self

    @ensure_fit
    def transform(self, X: ARRAY) -> ARRAY:
        return X

    def inverse_transform(self, X: ARRAY) -> ARRAY:
        return X



from typing import Literal, Sequence, Tuple, Unpack

from matplotlib.axes._axes import Axes
from numpy import ndarray
from pandas import DataFrame, Series

# Generic Types
ARRAY = ndarray | Series | Sequence[int | float]

# Colormap types
COLOR = str | Tuple[float, float, float]
COLORS_SEQ = Sequence[COLOR]
COLOR_FORMAT = Literal['hex', 'rgb', 'hls']

# Plot types
BW_TYPES = int | float | Literal['silverman', 'scott']
BOUNDS = Sequence[Tuple[float | int, float | int]]
AXES = Axes
